#!/usr/bin/perl -I/usr/sausalito/perl
# $Id: /usr/sausalito/sbin/easy-migrate.pl
#
# This is the main script of Easy Migrate.
#
# Easy Migrate is designed to easily migrate all or selected  Vsites, Users, 
# Emails, Cronjobs, DNS, MySQL users and databases as well as important system 
# information to another BlueOnyx 5209R or BlueOnyx 5210R server.
#
# For the migration we SCP our helper script /usr/sausalito/bin/em-helper.pl
# to the source server from which we are migrating. We then execute this
# helper script via SSH remote command execution and let it report back all 
# relevant data for the migration.
#
# Settings are copied over as JSON encoded Hashes and actual data we want to
# transfer is pulled individually by RSYNC over SSH.
#
# You can also use this script for more than simple migrations. You can also 
# use it to keep two Vsites (or whole servers) in synchronization. For example
# if you want to run load balancing or otherwise need multiple instances of 
# the same Vsite(s).
#
# Easy Migrate will keep the Vsite group identical on migration. If a Vsite was
# "site22" on the source server and the system group "site22" is free on the
# target server, then the Vsite will be "site22" on the target as well.
#
# As Easy Migrate uses RSYNC to copy Vsites and Users repeat runs of the
# migration script will only copy the differences and CODB settings (if 
# changed) will be kept updated.
#
# A lot of work went into "Easy Migrate" and I hope you find it much more
# to your liking than our old kludge CMU. :-)
#

my $version = "014";

#
### Load required Perl modules:
#

use Getopt::Long;
use POSIX;
use Data::Dumper;
use Net::OpenSSH;
use List::Util 'first';
use Array::Utils qw(:all);
use Unix::PasswdFile;
use DBI;

# Debugging switch (0|1|2):
# 0 = off
# 1 = log to syslog
# 2 = log to screen
#
$DEBUG = "2";
if ($DEBUG) {
    if ($DEBUG eq "1") {
        use Sys::Syslog qw( :DEFAULT setlogsock);
    }
}

my $cceinit = '0';
my $host = hostname();
my $header_sent = '0';

# Show header:
&header;

#
### Check if we are 'root':
#
&root_check;

#
### Init:
#

# May contain data from source server:
my %EMDATA = ();
my %SYSTEM = ();
my %VSITES = ();
my %USERS = ();

# Contains baseline of target server:
my %TARGETSYSTEM = ();
my %TARGETVSITES = ();
my %TARGETUSERS = ();

#
### Command line option handling
#

my $stepping_Check = 0;
my $stepping_List = 0;
my $stepping_Migrate = 0;
my $stepping_Services = 0;
my $stepping_DNSonly = 0;
my $option_noDNS = 0;
my $stepping_MYSQLonly = 0;

my $source = '';
my $port = '22';
my $check = '0';
my $ssh = '0';
my $import_ip = '127.0.0.1';
my $import_ipv6 = '';
my @selected_vsites = ();
my @selected_early_vsites = ();
my @selected_late_vsites = ();
my $beeline = '';

GetOptions(
  'source=s'    => \$source,
  'port=i'      => \$port,
  'ip=s'        => \$import_ip,
  'ipv6=s'      => \$import_ipv6,
  'vsites=s@'   => \@selected_vsites,
  'check'       => \$stepping_Check,
  'list'        => \$stepping_List,
  'migrate'     => \$stepping_Migrate,
  'config'      => \$stepping_Services,
  'nodns'       => \$option_noDNS,
  'dnsonly'     => \$stepping_DNSonly,
  'mysqlonly'   => \$stepping_MYSQLonly,
  'version'     => sub { &GetVersion() },
  'help|h|?'    => sub { help() },
) or &help();


# Sense check:
if ($source eq '') {
    &help("Please specify some command line options.");
}

if ($stepping_MYSQLonly == '1') {
    &Check($source, $port);
    &GetBaseLine();
    &MigrateMySQL($source, $port);
    print "\nDone!\n";
    exit(0);
}

if ($stepping_DNSonly == '1') {
    &Check($source, $port);
    &GetBaseLine();
    &MigrateDNS();
    print "\nDone!\n";
    exit(0);
}

$narr = scalar(@selected_vsites);
if ($narr == 0) {
    @selected_vsites = ('all');
    $beeline = 'all Vsites';
    if ($source ne '') {
        $stepping_Migrate = 1;
    }
}
else {
    @newSelVsites = ();
    $beeline = "Vsite";
    $blmp = 0;
    foreach my $x (@selected_vsites) {
        if ($x =~ /,/ ){
            $blmp++;
            my @nparts = split(',', $x);
            foreach my $z (@nparts) {
                push @newSelVsites, $z;
                if ($blmp eq 1) {
                    $beeline .= "s $z";
                }
                else {
                    $beeline .= " $z";
                }
            }
            @selected_vsites = @newSelVsites;
        }
        if ($blmp eq 0) {
            $beeline .= " $x";
        }
    }
    if ($source ne '') {
        $stepping_Migrate = 1;
    }
}

# Always run the connection check:
if ($stepping_Check == 1) {
    &Check($source, $port);
    print "Done!\n";
    exit(0);
}

if ($stepping_List == 1) {
    &Check($source, $port);
    &ListVsites();
    print "\nDone!\n";
    exit(0);
}

if ($stepping_Services == 1) {
    &Check($source, $port);
    if ($cceinit == '0') {
        &BlueOnyxInit();
    }
    &SetServices($source, $port);
    print "\nDone!\n";
    exit(0);
}

if ($stepping_Migrate == 1) {
    &Check($source, $port);
    &GetBaseLine();
    &Migrate($source, $port);
}

#
### Fin:
#

print "Done!\n";

if ($cceinit == '1') {
    $cce->bye("SUCCESS");    
}
exit(0);

#
### Subs:
#

sub GetVersion {
    print "[VERSION] " . $version . "\n\n";
}

sub BlueOnyxInit {

    use CCE;
    use Base::HomeDir qw(homedir_get_group_dir homedir_get_user_dir);
    use Sys::Hostname;
    use JSON;

    #
    ### CCE:
    #

    use CCE;
    $cce = new CCE;
    $cce->connectuds();
    $cceinit = '1';
}

sub Check {

    $source = shift;
    $port = shift;
    $host = $source . ':' . $port;

    if ($source eq "") {
        &help("[SOURCE_MISSING] Please specify the IP or Hostname of the migration source via --source. This must be done as first switch.");
    }

    print "# Checking connection to Source Platform: $source\n\n";

    $ssh = Net::OpenSSH->new($host, timeout => 30000);
    if ($ssh->error ne '0') {
        &help("[SSH_FAIL] Unable to establish an SSH connection to $source:$port. Please make sure the IP/Hostname and Port are correct and that SSH logins without password are possible.\n Net::OpenSSH: " . Dumper($ssh->error) . "");
    }
    else {
        print "[SSH_OK] SSH connection successfully established.\n\n";
        if (!$ssh->test("/usr/sausalito/bin/em-helper.pl")) {
            # File /usr/sausalito/bin/em-helper.pl doesn't exist on source.
            $ssh->scp_put("/usr/sausalito/bin/em-helper.pl", "/usr/sausalito/bin/em-helper.pl") or die "scp failed: " . $ssh->error;
            $ssh->system("chmod 755 /usr/sausalito/bin/em-helper.pl") or die "remote command failed: " . $ssh->error;
        }
    }

    # Now run the command:
    my $CMD = $ssh->capture("/usr/sausalito/bin/em-helper.pl --check");
    @ret = parser('\[SOURCE_OK\]', $CMD);
    if ($ret[0] eq "1") {
        print $ret[1] . "\n\n";
    }
    else {
        &help("$ret[1]\n");
    }
}

sub ListVsites {

    # Now run the command:
    my $CMD = $ssh->capture("/usr/sausalito/bin/em-helper.pl --list");
    print $CMD . "\n";

    if ($cceinit == '1') {
        $cce->bye("SUCCESS");    
    }
    exit(0);
}

sub SetServices {
    $source = shift;
    $port = shift;
    $host = $source . ':' . $port;

    print "[SERVICE_CONFIGURATION] Configuring services and other settings on this server identically to the source server.\n\n";

    # Get System Object from Source:
    my $CMD = $ssh->capture("/usr/sausalito/bin/em-helper.pl --system");
    @ret = parser('\[SYSTEM\]', $CMD, 1);
    if ($ret[0] eq "1") {
        $SYSTEM = $ret[1];
    }
    else {
        &help("$ret[1]\n");
    }

    # Which services 'System' NameSpaces do we care about is defined here:
    @SourceSystemServices = ('Sitestats', 'DNS', 'subdomains', 'Snmp', 'Ftp', 'VsiteDefaults', 'UserDefaults', 'Telnet', 'Email', 'MYSQLUSERS_DEFAULTS', 'Remote', 'Support', 'Nginx', 'Web', 'yum');

    # Which other CODB objects we care about is defined here:
    @SourceExtraServices = ('ActiveMonitor', 'PHP');

    foreach my $S_NS (keys %{ $SYSTEM->{'System'}->{'NameSpaces'} }) {
        if (in_array(\@SourceSystemServices, $S_NS)) {
            print "[INFO] Updating settings of 'System' NameSpace '$S_NS'.\n";
            $ObjectData = $SYSTEM->{'System'}->{'NameSpaces'}->{$S_NS};
            if ($S_NS eq 'Sitestats') {
                # This NameSpace requires some cleanup:
                delete $ObjectData->{'update'};
                $ret = &CCE_Tramp('System', $S_NS, 'none', $ObjectData);
                if ($ret == '-1') {
                    print "[WARN] The 'System' NameSpace settings of service '$S_NS' could not be set properly. Please review the settings manually.\n";
                }
            }
            elsif ($S_NS eq 'Support') {
                # This NameSpace requires some cleanup:
                delete $ObjectData->{'sos_delete'};
                delete $ObjectData->{'bugreport_trigger'};
                delete $ObjectData->{'sos_generate'};
                delete $ObjectData->{'sos_trigger'};
                delete $ObjectData->{'access_epoch'};
                delete $ObjectData->{'sos_epoch'};
                delete $ObjectData->{'ticket_trigger'};
                delete $ObjectData->{'support_account_name'};
                delete $ObjectData->{'support_account'};
                if ($ObjectData->{'client_email'} ne '') {
                    $ret = &CCE_Tramp('System', $S_NS, 'none', $ObjectData);
                    if ($ret == '-1') {
                        print "[WARN] The 'System' NameSpace settings of service '$S_NS' could not be set properly. Please review the settings manually.\n";
                    }                    
                }
                else {
                    print "[INFO] The 'System' NameSpace settings of service '$S_NS' are not being updated, because the client email address was empty on the source server.\n";
                }
            }
            elsif ($S_NS eq 'yum') {
                # This NameSpace requires some cleanup:
                delete $ObjectData->{'check'};
                delete $ObjectData->{'y_force_update'};
                delete $ObjectData->{'skiplock'};
                $ret = &CCE_Tramp('System', $S_NS, 'none', $ObjectData);
                if ($ret == '-1') {
                    print "[WARN] The 'System' NameSpace settings of service '$S_NS' could not be set properly. Please review the settings manually.\n";
                }
            }
            elsif ($S_NS eq 'VsiteDefaults') {
                # This NameSpace requires some cleanup:
                $ObjectData->{'ipaddr'} = $import_ip;
                $ret = &CCE_Tramp('System', $S_NS, 'none', $ObjectData);
                if ($ret == '-1') {
                    print "[WARN] The 'System' NameSpace settings of service '$S_NS' could not be set properly. Please review the settings manually.\n";
                }
            }
            else {
                # The rest we can set as is:
                $ret = &CCE_Tramp('System', $S_NS, 'none', $ObjectData);
                if ($ret == '-1') {
                    print "[WARN] The 'System' NameSpace settings of service '$S_NS' could not be set properly. Please review the settings manually.\n";
                }
            }
        }
    }
    print "\n";

    # ActiveMonitor:
    if ($SYSTEM->{'System'}->{'ActiveMonitor'}) {
        print "[INFO] Updating settings of 'ActiveMonitor'.\n";
        $ObjectData = $SYSTEM->{'System'}->{'ActiveMonitor'};
        delete $ObjectData->{'globalState'};
        $ret = &CCE_Tramp('ActiveMonitor', '', 'none', $ObjectData);
        if ($ret == '-1') {
            print "[WARN] The settings of service 'ActiveMonitor' could not be set properly. Please review the settings manually.\n";
        }
    }

    # PHP:
    if ($SYSTEM->{'System'}->{'PHP'}) {
        print "[INFO] Updating settings of 'PHP'.\n";
        $ObjectData = $SYSTEM->{'System'}->{'PHP'};
        delete $ObjectData->{'PHP_version_os'};
        delete $ObjectData->{'PHP_version'};
        delete $ObjectData->{'register_globals'};
        delete $ObjectData->{'register_globals_exception'};
        delete $ObjectData->{'php_ini_location'};
        delete $ObjectData->{'applicable'};
        $ret = &CCE_Tramp('PHP', '', 'none', $ObjectData);
        if ($ret == '-1') {
            print "[WARN] The settings of service 'PHP' could not be set properly. Please review the settings manually.\n";
        }
    }
    print "\n";
}

sub Migrate {

    $source = shift;
    $port = shift;
    $host = $source . ':' . $port;

    print "[MIGRATION_START] Starting Migration of $beeline\n";

    # Get System Object from Source:
    my $CMD = $ssh->capture("/usr/sausalito/bin/em-helper.pl --system");
    @ret = parser('\[SYSTEM\]', $CMD, 1);
    if ($ret[0] eq "1") {
        $SYSTEM = $ret[1];
    }
    else {
        &help("Error during run of '/usr/sausalito/bin/em-helper.pl --system' on source server. $ret[1]\nSSH error: " . Dumper($ssh->error) . "\n");
    }

    # Get Vsite Objects from Source:
    my $CMD = $ssh->capture("/usr/sausalito/bin/em-helper.pl --vsites");
    @ret = parser('\[VSITES\]', $CMD, 1);
    if ($ret[0] eq "1") {
        $VSITES = $ret[1];
    }
    else {
        &help("Error during run of '/usr/sausalito/bin/em-helper.pl --vsites' on source server. $ret[1]\nSSH error: " . Dumper($ssh->error) . "\n");
    }

    # Get User Objects from Source:
    my $CMD = $ssh->capture("/usr/sausalito/bin/em-helper.pl --users");
    @ret = parser('\[USERS\]', $CMD, 1);
    if ($ret[0] eq "1") {
        $USERS = $ret[1];
    }
    else {
        &help("Error during run of '/usr/sausalito/bin/em-helper.pl --users' on source server. $ret[1]\nSSH error: " . Dumper($ssh->error) . "\n");
    }

    #
    ### Check which Vsite groups are already taken on this server:
    #

    # First step: See which Vsites this server already has:

    my %TARGET_TAKEN_GROUPS = ();
    (@taken_Vsites_OIDS) = $cce->find('Vsite');
    foreach my $oidV (@taken_Vsites_OIDS) {
        ($ok, $TAKEN_VSITE_OBJ_DATA) = $cce->get($oidV);
        if ($ok) {
            $TARGET_TAKEN_GROUPS{'by_name'}{$TAKEN_VSITE_OBJ_DATA->{'name'}} = $TAKEN_VSITE_OBJ_DATA->{'fqdn'};
            $TARGET_TAKEN_GROUPS{'by_fqdn'}{$TAKEN_VSITE_OBJ_DATA->{'fqdn'}} = $TAKEN_VSITE_OBJ_DATA->{'name'};
        }
    }

    # Second step: Check existing Vsites on *this* server against the ones from the import:
    my %VSITES_THAT_NEED_NEW_GROUP = ();
    foreach my $name (keys %{ $TARGET_TAKEN_GROUPS{'by_name'} }) {
        if (defined($VSITES->{'NAME_INDEX'}->{$name})) {
            # Vsite waiting for import wants a group name of an existing Vsite. Check if FQDN is identical:
            if ($VSITES->{'NAME_INDEX'}->{$name} eq $TARGET_TAKEN_GROUPS{'by_name'}{$name}) {
                # Import contains Vsite with same group name and FQDN as already exists on this server.
                # Which is fine. We'll just sync them.
            }
            else {
                # Ok, we found a colission: Vsite in the import has a group name that is already taken here
                # by a Vsite with a *different* FQDN. So we take note of the Vsite from the import and store 
                # its information in the HASH %VSITES_THAT_NEED_NEW_GROUP. We'll then import those last and 
                # let them inherit a new group name.
                $VSITES_THAT_NEED_NEW_GROUP{'by_name'}{$name} = $VSITES->{'NAME_INDEX'}->{$name};
                $VSITES_THAT_NEED_NEW_GROUP{'by_fqdn'}{$VSITES->{'NAME_INDEX'}->{$name}} = $name;
            }
        }
    }

    #
    ### Sense Check for Vsites: What to import when.
    #

    # This could be done by fewer lines of code, but for readability this will do better:
    #
    # Sort through all Vsites destined for migration and sort them into two different
    # arrays: 
    #
    # @selected_early_vsites: Vsites w/o existing group conflicts (new Vsite or SYNC of existing Vsite)
    # @selected_late_vsites:  Vsites *with* group conflicts that we import last and where we assign new
    # groups.
    #

    @selected_early_vsites = ();
    @selected_late_vsites = ();
    if ($selected_vsites[0] eq "all") {
        # As we do not yet know what 'all' means, we loop through all known Vsites from the import
        # and check if they have conflicts. Depending on that we add them to one of the two arrays.
        foreach my $V_Group (keys %{ $VSITES->{'NAME_INDEX'} }) {
            if ($VSITES_THAT_NEED_NEW_GROUP{'by_name'}{$V_Group}) {
                push @selected_late_vsites, $VSITES->{'NAME_INDEX'}->{$V_Group};
            }
            else {
                push @selected_early_vsites, $VSITES->{'NAME_INDEX'}->{$V_Group};
            }
        }
    }
    else {
        # We know which Vsites should be imported. Check if they have conflicts and according to
        # that add them to the respective import arrays:
        foreach my $FQDN (@selected_vsites) {
            if ($VSITES_THAT_NEED_NEW_GROUP{'by_fqdn'}{$FQDN}) {
                push @selected_late_vsites, $VSITES->{'NAME_INDEX'}->{$VSITES->{'INDEX'}->{$FQDN}};
            }
            else {
                push @selected_early_vsites, $VSITES->{'NAME_INDEX'}->{$VSITES->{'INDEX'}->{$FQDN}};
            }
        }
    }

    #
    ### Check for fat fingers and if a Vsite specified for migration doesn't exist on source:
    #

    if ($selected_vsites[0] ne "all") {
        $numVsites = scalar(@selected_vsites);
        $foundVsites = 0;
        $beebugline = '';
        foreach my $FQDN (@selected_vsites) {
            if ($VSITES->{'INDEX'}{$FQDN} =~ /^site(.*)$/ ) {
                $foundVsites++;
            }
            else {
                $beebugline .= $FQDN . " ";
            }
        }

        # Something doesn't add up. Vsites were requested for Migration that don't exist on source server:
        if ($numVsites ne $foundVsites) {
            &help("The following Vsites requested for migration were not found on the source server: $beebugline");
        }
    }

    if ((scalar(@selected_early_vsites)) gt '0') {
        &doImportVsite('EARLY');
    }
    if ((scalar(@selected_late_vsites)) gt '0') {
        &doImportVsite('LATE');
    }

    #
    ### Fin
    #
}

sub doImportVsite {

    my $IMPORT_TIMER = shift;

    if (($IMPORT_TIMER ne 'EARLY') && ($IMPORT_TIMER ne 'LATE')) {
        &help("Unable to determine EARLY/LATE Import timer.");
    }
    elsif ($IMPORT_TIMER eq 'EARLY') {
        @vsites_to_import = @selected_early_vsites;
    }
    else {
        @vsites_to_import = @selected_late_vsites;
    }

    if (scalar(@vsites_to_import) eq '0') {
        &help("No Vsites specified for import!");
    }

    # Get Baseline of Vsites from our TARGET INDEX:
    my $mk = '';
    foreach my $key (keys %{ $TARGETVSITES{'INDEX'} }) {
        if ($mk eq '') {
            $FQDN = $key;
            $SiteName = $TARGETVSITES{'INDEX'}{$key};
        }
    }
    $TARGET_VSITE_BASELINE = GetBaselineObject($SiteName, 'TARGETVSITES', 'Vsite');
    if ($TARGET_VSITE_BASELINE eq '-1') {
        &help("Unable to get target Vsite baseline.");   
    }

    #
    ### Compare 
    #

    @DUMMY_VSITE_OIDS = ();

    #print Dumper(\@DUMMY_VSITE_OIDS);
    foreach my $FQDN (@vsites_to_import) {

        undef $NEW_VSITE_OBJ_DATA;

        $dummyVsiteNum = '1';
        if ($FQDN eq 'n/a') {
                $cce->create("Vsite", { 'fqdn' => "$dummyVsiteNum.blueonyx.it", 'hostname' => "$dummyVsiteNum", 'createdUser' => 'admin', 'volume' => "/home", 'domain' => "blueonyx.it", 'ipaddr' => "127.0.0.1" });
                $DUMMY_VSITE_OIDS[$dummyVsiteNum] = $cce->oid();
                $dummyVsiteNum++;
        }

        # Get the first Vsite data from the source lined up:
        $Import_VSITE = GetObject($FQDN, 'VSITES', 'INDEX');

        %VsiteObject = ();
        $VsiteObject = $Import_VSITE->{'Vsite'};

        if ($Import_VSITE ne "-1") {

            # We got valid Vsite data
            @BaseLineKeys = (keys %$TARGET_VSITE_BASELINE);
            @VsiteKeys = (keys %$VsiteObject);

            # Get the difference between Vsite Baseline and import Vsite keys:
            my @diff = array_diff(@BaseLineKeys, @VsiteKeys);
            foreach my $x (@diff) {
                # If the import Vsite has keys that we do NOT have locally, then we delete them here:
                delete $VsiteObject->{$x};
            }

            # Get old group name of Import Vsite:
            $old_site_name = $VsiteObject->{'name'};
            $old_site_ipv4 = $VsiteObject->{'ipaddr'};
            $old_site_ipv6 = $VsiteObject->{'ipaddrIPv6'};

            #
            # Import Vsite Object Cleanup:
            #
            # Unconditionally delete the following keys:
            #
            # - OID
            # - CLASSVER
            # - CLASS
            # - basedir
            #

            # Store this for later for the directory Sync:
            $source_Vsite_Basedir = $VsiteObject->{'basedir'};

            delete $VsiteObject->{'OID'};
            delete $VsiteObject->{'CLASSVER'};
            delete $VsiteObject->{'CLASS'};
            delete $VsiteObject->{'basedir'};

            # Only delete the group on late imports of Vsites in order to auto-assign new group names:
            if ($IMPORT_TIMER eq "LATE") {
                delete $VsiteObject->{'name'};
            }

            # Fix the import IPv4 and IPv6 IPs:
            $VsiteObject->{'ipaddr'} = $import_ip;
            $VsiteObject->{'ipaddrIPv6'} = $import_ipv6;

            # Value for 'maxusers' my be empty bool. Fix that to the default if it is:
            if ($VsiteObject->{'maxusers'} eq "") {
                $VsiteObject->{'maxusers'} = '25';
            }

            #
            ### Check if this Vsite is owned by a reseller:
            #

            if (($VsiteObject->{'createdUser'} ne 'admin') && (!$TARGETUSERS->{'users'}->{$VsiteObject->{'createdUser'}})) {
                print "[INFO] Vsite $VsiteObject->{'fqdn'} is owned by an Extra-Admin that's not yet aboard. Fixing that.\n";
                if (!$USERS->{'users'}->{$VsiteObject->{'createdUser'}}) {
                    &help("Vsite $VsiteObject->{'fqdn'} is owned by Reseller '$USERS->{'users'}->{$VsiteObject->{'createdUser'}}', but that user does not exist on the source server.");
                }
                else {
                    # Import the Extra-Admin and his files:
                    &UserImport({ $VsiteObject->{'createdUser'} => $USERS->{'users'}->{$VsiteObject->{'createdUser'}}}, '');
                    &UserFileImport({ $VsiteObject->{'createdUser'} => $USERS->{'users'}->{$VsiteObject->{'createdUser'}}}, 'users');
                }
            }

            #
            ### Check if this Vsite is owned by a reseller and if that Reseller has access to the IP the Vsite is supposed to use:
            #

            if (($VsiteObject->{'createdUser'} ne 'admin') && ($TARGETUSERS->{'users'}->{$VsiteObject->{'createdUser'}})) {
                # Only bother if Network pooling is enabled:
                if ($Network->{'pooling'} eq '1') {
                    # Now make sure that reseller is allowed to use the IP we're assigning to the Vsite:
                    if ($import_ip ne '') {
                        (@POOL_OIDS) = $cce->find('IPPoolingRange', { 'min' => $import_ip, 'max' => $import_ip });
                        if (scalar(@POOL_OIDS)) {
                            ($ok, $IPPoolingRange) = $cce->get($POOL_OIDS[0]);
                            if ($ok) {
                                @pool_admins = $cce->scalar_to_array($IPPoolingRange->{'admin'});
                                if (in_array(\@pool_admins, $VsiteObject->{'createdUser'})) {
                                    # nada
                                }
                                else {
                                    push @pool_admins, $VsiteObject->{'createdUser'};
                                    $new_pool_admins = $cce->array_to_scalar(@pool_admins);
                                    ($ok) = $cce->set($POOL_OIDS[0], '', { 'admin' => $new_pool_admins });
                                }
                            }
                        }
                    }
                    if ($import_ipv6 ne '') {
                        (@POOL_OIDS) = $cce->find('IPPoolingRange', { 'min' => $import_ipv6, 'max' => $import_ipv6 });
                        if (scalar(@POOL_OIDS)) {
                            ($ok, $IPPoolingRange) = $cce->get($POOL_OIDS[0]);
                            if ($ok) {
                                @pool_admins = $cce->scalar_to_array($IPPoolingRange->{'admin'});
                                if (in_array(\@pool_admins, $VsiteObject->{'createdUser'})) {
                                    # nada
                                }
                                else {
                                    push @pool_admins, $VsiteObject->{'createdUser'};
                                    $new_pool_admins = $cce->array_to_scalar(@pool_admins);
                                    ($ok) = $cce->set($POOL_OIDS[0], '', { 'admin' => $new_pool_admins });
                                }
                            }
                        }
                    }
                }
            }

            #
            ### Actual Vsite creation & population:
            #

            my @VSITE_OIDS = ();
            my $Compass_webapps_Vsite_SQL_User_Disable_File = "/tmp/Compass_webapps-" . $VsiteObject->{'name'};
            (@VSITE_OIDS) = $cce->find('Vsite', { 'fqdn' => $VsiteObject->{'fqdn'} });

            # Create lockfile to let the Handler vsite_autoinstall.pl of Compass WebApps know to do nothing:
            $Compass_webapps_Vsite_SQL_User_Disable_File = "/tmp/Compass_webapps-" . $VsiteObject->{'name'};
            system("touch $Compass_webapps_Vsite_SQL_User_Disable_File");

            if ((scalar(@VSITE_OIDS)) eq '0') {
                # No Vsite with this FQDN was found. Good. 

                # Check if /etc/group has a group of said name. It shouldn't:
                $gcheck = $VsiteObject->{'name'};
                $groupcheck = `cat /etc/group|grep "$gcheck:"|wc -l`;
                chomp($groupcheck);
                if ($groupcheck eq '1') {
                    # For some reason we don't have a Vsite of this name, but we have a group for it. Delete the group:
                    print "\n[INFO] /etc/group already has a group called $gcheck, but no matching Vsite. Deleting group $gcheck.\n";
                    system("/usr/sbin/groupdel $gcheck");
                }

                # Create Vsite:
                $cce->create("Vsite", $VsiteObject);
                $VSITE_OIDS[0] = $cce->oid();

                # Make sure it's there:
                ($ok, $NEW_VSITE_OBJ_DATA) = $cce->get($VSITE_OIDS[0]);
                if (!$ok) {
                    # Remove temporary lockfile for the handler to trip on:
                    system("rm -f $Compass_webapps_Vsite_SQL_User_Disable_File");
                    &help("Vsite $VsiteObject->{'fqdn'} did not create properly.");
                }
                else {
                    print "[INFO] Vsite $VsiteObject->{'fqdn'} created.\n";
                }
            }
            else {
                # Vsite is already present. Update it with the new settings from the import:
                ($ok) = $cce->set($VSITE_OIDS[0], '', $VsiteObject);
                if (!$ok) {
                    &help("Failed to update settings of Vsite $VsiteObject->{'fqdn'}");
                }
                else {
                    print "\n[INFO] Settings of Vsite $VsiteObject->{'fqdn'} updated.\n";
                }
            }

            # Directly get the new Vsite Object:
            if (!defined($NEW_VSITE_OBJ_DATA)) {
                ($ok, $NEW_VSITE_OBJ_DATA) = $cce->get($VSITE_OIDS[0]);
            }

            #
            ### Populate Vsite NameSpaces:
            #

            @missing_vs_namespaces = ();
            $TARGET_VSNS_BASELINE = GetBaselineObject($SiteName, 'TARGETVSITES', 'NameSpaces');
            foreach my $key (keys %{ $Import_VSITE->{'NameSpaces'} }) {
                %Vsite_NameSpace = ();
                $Vsite_NameSpace = $Import_VSITE->{'NameSpaces'}->{$key};
                if (!$TARGET_VSNS_BASELINE->{$key}) {
                    #print "[INFO] Vsite NameSpace '$key' doesn't exist. You might not have the PKG for that installed. Ignoring settings on import.\n";
                    push @missing_vs_namespaces, $key;
                    delete $Import_VSITE->{'NameSpaces'}->{$key}
                }
            }

            # Get the cleaned results again into a Hash:
            %Vsite_NameSpace = ();
            $Vsite_NameSpace = $Import_VSITE->{'NameSpaces'};

            # Set up Hash of NameSpaces that we must run delayed:
            %Vsite_NameSpace_Delayed = ();
            @delayed_vsite_namespaces = ('SSL');

            $vsns_infoline = '0';
            foreach my $key (keys %{ $Vsite_NameSpace }) {
                %Vsite_NameSpace = ();
                $Vsite_NameSpace = $Import_VSITE->{'NameSpaces'}->{$key};

                %Baseline_NameSpace = ();
                $Baseline_NameSpace = $TARGET_VSNS_BASELINE->{$key};

                # We got valid Vsite data
                my @BaseLineKeys = (keys %$Baseline_NameSpace);
                my @VsiteKeys = (keys %$Vsite_NameSpace);

                # Get the difference between Vsite Baseline and import Vsite keys:
                my @diff = array_diff(@BaseLineKeys, @VsiteKeys);
                foreach my $x (@diff) {
                    # If the import Vsite has NameSpace keys that we do NOT have locally, then we delete them here:
                    delete $Vsite_NameSpace->{$x};
                    #print "Diff in $key: $x \n"
                }

                # Cleanup:
                delete $Vsite_NameSpace->{'NAMESPACE'};
                delete $Vsite_NameSpace->{'CLASSVER'};

                #
                # Sense Checks: Some NameSpace's don't set with empty values:
                #

                if (($key eq 'Disk') && ($Vsite_NameSpace->{'used'} eq "")) {
                    $Vsite_NameSpace->{'used'} = '0';
                }
                if ($key eq 'SSL') {
                    if ($Vsite_NameSpace->{'country'} eq "") {
                        $Vsite_NameSpace->{'country'} = 'US';
                    }
                    delete $Vsite_NameSpace->{'LEclientRet'};
                    delete $Vsite_NameSpace->{'performLEinstall'};
                    delete $Vsite_NameSpace->{'importCert'};
                    delete $Vsite_NameSpace->{'createCert'};
                    delete $Vsite_NameSpace->{'createCsr'};
                    delete $Vsite_NameSpace->{'caCerts'};

                    # Set up 'SSL' NameSpace to run delayed after we're done with file imports:
                    $Vsite_NameSpace_Delayed->{'SSL'} = $Vsite_NameSpace;
                    next;
                }
                if ($key eq 'MYSQL_Vsite') {
                    if (defined($Vsite_NameSpace->{'doBackupDB'})) {
                        delete $Vsite_NameSpace->{'doBackupDB'};
                        delete $Vsite_NameSpace->{'doBackupDBname'};
                        delete $Vsite_NameSpace->{'doRestoreDB'};
                        delete $Vsite_NameSpace->{'doRestoreDBname'};
                    }
                    if (defined($Vsite_NameSpace->{'host'})) {
                        if ($Vsite_NameSpace->{'host'} eq 'localhost') {
                            $Vsite_NameSpace->{'host'} = '127.0.0.1';
                        }
                    }
                }

                if (($key eq 'MYSQL_Vsite') && (($Vsite_NameSpace->{'enabled'} eq "0") ||($Vsite_NameSpace->{'username'} eq "") || ($Vsite_NameSpace->{'pass'} eq "") || ($Vsite_NameSpace->{'DB'} eq ""))) {
                    # Ignore setting 'MYSQL_Vsite' as source Vsite wasn't using MysQL.
                    next;
                }

                if ($key eq 'PHP') {
                    if (defined $Vsite_NameSpace->{'version'}) {
                        if ($Vsite_NameSpace->{'version'} ne "PHPOS") {
                            # PHP is not using the OS supplied version. Danger, Will Robinson. Check if the PHP is installed and present:
                            $reported_PHP_short_Version = $Vsite_NameSpace->{'version'};
                            ($PHP_oid) = $cce->find('PHP');
                            ($ok, $Server_PHP) = $cce->get($PHP_oid);
                            ($ok, $Server_PHPPKG) = $cce->get($PHP_oid, $reported_PHP_short_Version);
                            if (($Server_PHPPKG->{'present'} == "1") && ($Server_PHPPKG->{'enabled'} == "1")) {
                                # Desired PHP version IS available. Use it:
                                $Vsite_NameSpace->{'version'} = $requested_PHP_short;
                            }
                            else {
                                # Desired PHP version not available. Default to 'PHPOS':
                                $Vsite_NameSpace->{'version'} = 'PHPOS';
                            }
                        }
                    }
                    else {
                        my $requested_PHP_long = $SYSTEM->{'System'}->{'OSPHP'};
                        my (@PHPhelper) = split /\./, $requested_PHP_long;
                        my $requested_PHP_short = "PHP" . $PHPhelper[0] . $PHPhelper[1];
                        # Find out what PHP we're using or have available:
                        ($PHP_oid) = $cce->find('PHP');
                        ($ok, $Server_PHP) = $cce->get($PHP_oid);
                        ($ok, $Server_PHPPKG) = $cce->get($PHP_oid, $requested_PHP_short);
                        if (($Server_PHPPKG->{'present'} == "1") && ($Server_PHPPKG->{'enabled'} == "1")) {
                            # Desired PHP version IS available. Use it:
                            $Vsite_NameSpace->{'version'} = $requested_PHP_short;
                        }
                        else {
                            # Desired PHP version not available. Default to 'PHPOS':
                            $Vsite_NameSpace->{'version'} = 'PHPOS';
                        }
                    }

                    # Handle difference in keys when we come from anything older than 5209R:
                    if (($SYSTEM{'System'}{'productBuild'} ne '5210R') && ($SYSTEM{'System'}{'productBuild'} ne '5209R')) {
                        if ($Vsite_NameSpace->{'suPHP_enabled'} eq "1") {
                            $Vsite_NameSpace->{'mod_ruid_enabled'} = "0";
                            $Vsite_NameSpace->{'fpm_enabled'} = "0";
                            $Vsite_NameSpace->{'enabled'} = "1";
                        }
                        elsif ($Vsite_NameSpace->{'enabled'} eq "1") {
                            $Vsite_NameSpace->{'suPHP_enabled'} = "1";
                            $Vsite_NameSpace->{'mod_ruid_enabled'} = "0";
                            $Vsite_NameSpace->{'fpm_enabled'} = "0";
                        }
                        elsif ($Vsite_NameSpace->{'mod_ruid_enabled'} eq "1") {
                            $Vsite_NameSpace->{'suPHP_enabled'} = "0";
                            $Vsite_NameSpace->{'fpm_enabled'} = "0";
                            $Vsite_NameSpace->{'enabled'} = "1";
                        }
                        elsif ($Vsite_NameSpace->{'fpm_enabled'} eq "1") {
                            $Vsite_NameSpace->{'suPHP_enabled'} = "0";
                            $Vsite_NameSpace->{'mod_ruid_enabled'} = "0";
                            $Vsite_NameSpace->{'enabled'} = "1";
                        }
                        else {
                            $Vsite_NameSpace->{'suPHP_enabled'} = "0";
                            $Vsite_NameSpace->{'mod_ruid_enabled'} = "0";
                            $Vsite_NameSpace->{'fpm_enabled'} = "0";
                            $Vsite_NameSpace->{'enabled'} = "0";
                        }
                    }
                }

                # Handle individual WebApps:
                if (($Vsite_NameSpace->{'appname'}) && ($Vsite_NameSpace->{'sqldb'}) && ($Vsite_NameSpace->{'revision'})) {
                    # Set up the WebApplication NameSpace to run delayed after we're done with importing NameSpace 'Compass_webapps':
                    $Vsite_NameSpace_Delayed->{$key} = $Vsite_NameSpace;
                    push @delayed_vsite_namespaces, $key;
                    next;
                }

                if ($key eq 'Compass_webapps') {
                    delete $Vsite_NameSpace->{'allowSiteAdmin'};
                    $Vsite_NameSpace->{'doNothing'} = time();
                    if ($Vsite_NameSpace->{'sql_pass'} eq '') {
                        delete$Vsite_NameSpace->{'sql_pass'};
                    }
                    if ($Vsite_NameSpace->{'sql_user'} eq '') {
                        delete$Vsite_NameSpace->{'sql_user'};
                    }
                }

                # Set NameSpace:
                ($ok) = $cce->set($VSITE_OIDS[0], $key, $Vsite_NameSpace);
                if (!$ok) {
                    &help("\n[ERROR] Failed to update settings of Vsite NameSpace '$key' of Vsite $VsiteObject->{'fqdn'}");
                }
                else {
                    #print "[INFO] Configured NameSpace '$key' of Vsite $VsiteObject->{'fqdn'}.\n";
                    if ($vsns_infoline == '0') {
                        print "[INFO] Configuring NameSpace '$key' ";
                        $vsns_infoline++;
                    }
                    else {
                        print "'$key' ";
                        $vsns_infoline++;
                    }
                }
            }

            # Remove Compass_webapps lockfile:
            system("rm -f $Compass_webapps_Vsite_SQL_User_Disable_File");

            print "\n\n";
            if (scalar(@missing_vs_namespaces) gt "0") {
                $outMsg = join(", ", @missing_vs_namespaces);
                print "[WARN] Several NameSpace's of this Vsite could not be set. You might not have the PKGs for these installed: $outMsg\n\n";
            }
            
            #
            ### Import Users of this Vsite:
            #

            $target_Vsite_Basedir = $NEW_VSITE_OBJ_DATA->{'basedir'};
            $new_vsite_group = $NEW_VSITE_OBJ_DATA->{'name'};

            &UserImport($USERS->{$old_site_name}, $new_vsite_group);

            #
            ### Import Vsite Files:
            #

            # Get basedir directories from source:
            undef %VSITE_SUBDIRS;
            undef %VSITE_DIROWNERS;
            undef @dirhelper;
            undef @dirhelper_precursor;
            undef $VS_ls;
            %$VSITE_SUBDIRS = ();
            %$VSITE_DIROWNERS = ();
            @dirhelper = ();
            @dirhelper_precursor = ();
            $VS_ls = '';
            #if ($SYSTEM{'System'}{'productBuild'} eq '5210R') {
            if ($SYSTEM->{'System'}->{'productBuild'} eq '5210R') {
                $VS_ls = $ssh->capture("find $source_Vsite_Basedir -maxdepth 2 -type d"); # Get directories only
                #print "Going two deep.\n";
                (@dirhelper_precursor) = split /\n/, $VS_ls;
                foreach my $x (@dirhelper_precursor) {
                    # Strip naked basedir:
                    if ($x eq $source_Vsite_Basedir) { next; }
                    $x =~ s{^$source_Vsite_Basedir/}{};
                    # Strip jail-directories:
                    if ($x =~ /^etc(.*)$/) { next; }
                    if ($x =~ /^usr(.*)$/) { next; }
                    if ($x =~ /^tmp(.*)$/) { next; }
                    if ($x =~ /^home(.*)$/) { next; }
                    if ($x =~ /^dev(.*)$/) { next; }
                    if ($x =~ /^var\/tmp$/) { next; }
                    if ($x =~ /^var\/tmp(.*)$/) { next; }
                    # Push the rest for later:
                    push @dirhelper, $x;
                }
            }
            else {
                #print "Going one deep.\n";
                $VS_ls = $ssh->capture("find $source_Vsite_Basedir -maxdepth 1 -type d"); # Get directories only
                (@dirhelper_precursor) = split /\n/, $VS_ls;
                foreach my $x (@dirhelper_precursor) {
                    if ($x eq $source_Vsite_Basedir) { next; }
                    $x =~ s{^.*/}{};
                    push @dirhelper, $x;
                }
            }

            # Get PHP Object of Vsite:
            ($ok, $NEW_VSITE_OBJ_PHP_DATA) = $cce->get($VSITE_OIDS[0], 'PHP');

            print "\n[INFO] Importing Files of Vsite $VsiteObject->{'fqdn'}\n";

            # Get default owners UIDs for directories:
            $uid_chown = 'root';
            if ($NEW_VSITE_OBJ_PHP_DATA->{'prefered_siteAdmin'} ne '') {
                $uid_chown = $NEW_VSITE_OBJ_PHP_DATA->{'prefered_siteAdmin'};
            }
            else {
                $uid_chown = 'nobody';
            }

            #print "dirhelper: " . Dumper(\@dirhelper);

            # Hash the directories for easier management and directly specify who owns them:
            foreach my $D (@dirhelper) {
                if (($D eq "users") || ($D eq ".users")) {
                    next;
                }
                if (($TARGETSYSTEM{'System'}{'productBuild'} eq '5210R') && ($SYSTEM->{'System'}->{'productBuild'} ne '5210R')) {
                    # Importing to 5210R from anything that's not a 5210R:
                    #print "# For 5210R target and non-5210R source adjust paths and set owners:\n";
                    if ($D eq 'awstats') {
                        $VSITE_SUBDIRS->{$D} = 'var/awstats';
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D eq 'logs') {
                        $VSITE_SUBDIRS->{$D} = 'var/logs';
                        $logUser = uc($new_vsite_group) . '-logs';
                        $VSITE_DIROWNERS->{$D} = "$logUser:$new_vsite_group";
                    }
                    elsif ($D eq 'webalizer') {
                        $VSITE_SUBDIRS->{$D} = 'var/webalizer';
                        $VSITE_DIROWNERS->{$D} = "httpd:$new_vsite_group";
                    }
                    elsif ($D eq 'certs') {
                        $VSITE_SUBDIRS->{$D} = 'wwwroot/certs';
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D eq 'php.d') {
                        $VSITE_SUBDIRS->{$D} = 'wwwroot/php.d';
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D eq 'webapps_backup') {
                        $VSITE_SUBDIRS->{$D} = 'wwwroot/webapps_backup';
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    else {
                        $VSITE_SUBDIRS->{$D} = "wwwroot/$D";
                        $VSITE_DIROWNERS->{$D} = "$uid_chown:$new_vsite_group";
                    }
                }
                elsif (($TARGETSYSTEM{'System'}{'productBuild'} eq '5210R') && ($SYSTEM->{'System'}->{'productBuild'} eq '5210R')) {
                    # Importing to 5210R from another 5210R:
                    #print "# For 5210R target and 5210R source keep existing paths and set owners:\n";
                    $VSITE_SUBDIRS->{$D} = $D;
                    if ($D =~ /^wwwroot\/web/) {
                        $VSITE_DIROWNERS->{$D} = "$uid_chown:$new_vsite_group";
                    }
                    elsif ($D =~ /^wwwroot\/php\.d/) {
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D =~ /^wwwroot\/webapps_backup/) {
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D =~ /^wwwroot\/certs/) {
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D =~ /^wwwroot(.*)$/) {
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D eq 'var/logs') {
                        $logUser = uc($new_vsite_group) . '-logs';
                        $VSITE_DIROWNERS->{$D} = "$logUser:$new_vsite_group";
                    }
                    else {
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                }
                elsif (($TARGETSYSTEM{'System'}{'productBuild'} eq '5209R') && ($SYSTEM->{'System'}->{'productBuild'} eq '5210R')) {
                    # We're going back from 5210R to 5209R:
                    #print "# We're going back from 5210R to 5209R:\n";
                    if (($D eq 'var') || ($D eq 'wwwroot')) {
                        next;
                    }
                    elsif ($D eq 'var/logs') {
                        $VSITE_SUBDIRS->{$D} = 'logs';
                        $logUser = uc($new_vsite_group) . '-logs';
                        $VSITE_DIROWNERS->{$D} = "$logUser:$new_vsite_group";
                    }
                    elsif ($D eq 'var/webalizer') {
                        $VSITE_SUBDIRS->{$D} = 'webalizer';
                        $VSITE_DIROWNERS->{$D} = "httpd:$new_vsite_group";
                    }
                    elsif ($D eq 'var/awstats') {
                        $VSITE_SUBDIRS->{$D} = 'awstats';
                        $VSITE_DIROWNERS->{$D} = "httpd:$new_vsite_group";
                    }
                    elsif ($D eq 'sql') {
                        $VSITE_SUBDIRS->{$D} = 'sql';
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D eq 'wwwroot/certs') {
                        $VSITE_SUBDIRS->{$D} = 'certs';
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D eq 'wwwroot/certs') {
                        $VSITE_SUBDIRS->{$D} = 'webapps_backup';
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D eq 'wwwroot/php.d') {
                        $VSITE_SUBDIRS->{$D} = 'webapps_backup';
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D eq 'wwwroot/webapps_backup') {
                        $VSITE_SUBDIRS->{$D} = 'webapps_backup';
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D =~ /^wwwroot(.*)$/) {
                        $VSITE_SUBDIRS->{$D} = (split '/', $D)[-1];
                        $VSITE_DIROWNERS->{$D} = "$uid_chown:$new_vsite_group";
                    }
                    else {
                        $VSITE_SUBDIRS->{$D} = (split '/', $D)[-1];
                        $VSITE_DIROWNERS->{$D} = "$uid_chown:$new_vsite_group";
                    }
                }
                else {
                    # For anything older than 5210R keep paths, but set owners:
                    #print "# For anything older than 5210R keep paths, but set owners:\n";
                    $VSITE_SUBDIRS->{$D} = $D;
                    if ($D eq 'awstats') {
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D eq 'logs') {
                        $logUser = uc($new_vsite_group) . '-logs';
                        $VSITE_DIROWNERS->{$D} = "$logUser:$new_vsite_group";
                    }
                    elsif ($D eq 'webalizer') {
                        $VSITE_DIROWNERS->{$D} = "apache:$new_vsite_group";
                    }
                    elsif ($D eq 'certs') {
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D eq 'php.d') {
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    elsif ($D eq 'webapps_backup') {
                        $VSITE_DIROWNERS->{$D} = "root:$new_vsite_group";
                    }
                    else {
                        $VSITE_DIROWNERS->{$D} = "$uid_chown:$new_vsite_group";
                    }
                }
            }

            #print "VSITE_SUBDIRS: " . Dumper(\$VSITE_SUBDIRS);
            #print "VSITE_DIROWNERS: " . Dumper(\$VSITE_DIROWNERS);

            # Start Rsync of Vsite data:
            foreach my $directory (keys %{ $VSITE_SUBDIRS }) {
                $sourceDir = $source_Vsite_Basedir . '/' . $directory . '/';
                $target_Vsite_actiondir = "$target_Vsite_Basedir/$VSITE_SUBDIRS->{$directory}/";

                # Create target directory if it doesn't exist:
                if (! -d $target_Vsite_actiondir) {
                    system("mkdir -p $target_Vsite_actiondir");
                }

                # Actual Rsync:
                $DirOwner = $VSITE_DIROWNERS->{$directory};
                print "[INFO] Starting RSYNC from $host:$sourceDir to $target_Vsite_actiondir\n";
                #print "/usr/bin/rsync -q --progress -ar '-e ssh -p $port' $source:$sourceDir $target_Vsite_actiondir --delete-after --info=progress2 --chown=$DirOwner\n";
                system("/usr/bin/rsync -q --progress -ar '-e ssh -p $port' $source:$sourceDir $target_Vsite_actiondir --delete-after --info=progress2 --chown=$DirOwner >/dev/null 2>&1 || :");
            }

            # Fix owner of sitedir:
            if ($TARGETSYSTEM{'System'}{'productBuild'} eq '5210R') {
                # Enforce jail compatible UID on $target_Vsite_Basedir;
                #print "chown root:root $target_Vsite_Basedir\n";
                system("chown root:root $target_Vsite_Basedir");
            }
            else {
                # Enforce proper ownership on $target_Vsite_Basedir:
                #print "chown $uid_chown:$new_vsite_group $target_Vsite_Basedir\n";
                system("chown $uid_chown:$new_vsite_group $target_Vsite_Basedir");
            }

            if ($TARGETSYSTEM{'System'}{'productBuild'} eq '5210R') {
                # On 5210R: Fix owner of sitedir/wwwroot to work with suPHP:
                #print "chown $uid_chown:$new_vsite_group $target_Vsite_Basedir/wwwroot\n";
                system("chown $uid_chown:$new_vsite_group $target_Vsite_Basedir/wwwroot");

                # Also on 5210R: Fix owner of var/logs to be site's loguser:
                $logUser = uc($new_vsite_group) . '-logs';
                #print "chown $logUser:$new_vsite_group $target_Vsite_Basedir/var/logs\n";
                system("chown $logUser:$new_vsite_group $target_Vsite_Basedir/var/logs");

                # Above we were ignoring a sync of the Jails /tmp directory. So it may be missing.
                # Here is where we create it empty and throw on ownerships and permissions:
                if ((-d "$target_Vsite_Basedir/var/tmp") && (!-d "$target_Vsite_Basedir/tmp")) {
                    system("mkdir $target_Vsite_Basedir/tmp");
                    system("chown root:root $target_Vsite_Basedir/tmp");
                    system("chmod 0777 $target_Vsite_Basedir/tmp");
                }
            }

            foreach my $key (keys %{ $Vsite_NameSpace_Delayed }) {
                # Special case for 'roundcubemail'. This needs 'LOCK_TABLES' privileges for successful DB import.
                # It didn't get it by default, so we hastily set it here:
                if ($Vsite_NameSpace_Delayed->{$key}->{'appname'} eq 'roundcubemail') {
                    print "[INFO] WebApp 'roundcubemail' present. Granting SQL user 'LOCK_TABLES' privileges for this.\n";
                    ($ok) = $cce->set($VSITE_OIDS[0], 'MYSQL_Vsite', {'LOCK_TABLES' => '1'});
                }

                # Handle individual WebApps:
                $webapps_backup_sql_file = "$target_Vsite_Basedir/$VSITE_SUBDIRS->{webapps_backup}/" . $Vsite_NameSpace_Delayed->{$key}->{'sqldb'} . '.sql';
                if (-f $webapps_backup_sql_file) {

                    print "[INFO] Importing '$key' SQL database '$Vsite_NameSpace_Delayed->{$key}->{'sqldb'}'\n";

                    $sql_query  = "\"DROP DATABASE IF EXISTS $Vsite_NameSpace_Delayed->{$key}->{'sqldb'}; \"";
                    $sql_query .= "\"CREATE DATABASE IF NOT EXISTS $Vsite_NameSpace_Delayed->{$key}->{'sqldb'}; \"";

                    ## MySQL Server Connection Check
                    $dbh = DBI->connect(
                            "DBI:mysql:mysql:$$TARGETSYSTEM{'System'}{'MySQL'}{'sql_host'}:$TARGETSYSTEM{'System'}{'MySQL'}{'sql_port'}",
                            $TARGETSYSTEM{'System'}{'MySQL'}{'sql_root'}, $TARGETSYSTEM{'System'}{'MySQL'}{'sql_rootpassword'},
                            {
                                RaiseError => 0,
                                PrintError => 0
                            }
                    ); 
                    if (!$dbh){
                        &help("Unable to connect to MySQL/MariaDB on the this server. Please make sure MySQL/MariaDB is running and that the connection details for MySQL are configured correctly in the GUI.");
                    }

                    # Drop Database
                    my $this_DB = $Vsite_NameSpace_Delayed->{$key}->{'sqldb'};
                    $query = "DROP DATABASE IF EXISTS `$this_DB`;\n";
                    $return = $dbh->do($query);
                    if ($dbh->err) {
                         print "[WARN] The following error happened while trying to import SQL database '$Vsite_NameSpace_Delayed->{$key}->{'sqldb'}': $dbh->errstr\n";
                    }

                    # Create Database
                    my $this_DB = $Vsite_NameSpace_Delayed->{$key}->{'sqldb'};
                    $query = "CREATE DATABASE IF NOT EXISTS `$this_DB`;\n";
                    $return = $dbh->do($query);
                    if ($dbh->err) {
                         print "[WARN] The following error happened while trying to import SQL database '$Vsite_NameSpace_Delayed->{$key}->{'sqldb'}': $dbh->errstr\n";
                    }

                    # Import SQL-Dump:
                    #print "Running: /usr/bin/mysql -u " . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_root'} . " -p" . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_rootpassword'} . " " . $this_DB . " < $webapps_backup_sql_file\n";
                    system("/usr/bin/mysql -u " . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_root'} . " -p" . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_rootpassword'} . " " . $this_DB . " < $webapps_backup_sql_file >/dev/null 2>&1 || :");
                }

                # Handle individual WebApps. Set 'doRestoreDB', but reset all other backup/restore related keys. Also set 'doNothing',
                # so that our handlers know to skip a beat as this is a restore and not a new deploy:
                if (($Vsite_NameSpace_Delayed->{$key}->{'appname'}) && ($Vsite_NameSpace_Delayed->{$key}->{'sqldb'}) && ($Vsite_NameSpace_Delayed->{$key}->{'revision'})) {
                    # Line up the ducks for the toplevel 'WebApplications' object, but also set doNothing:
                    $WA_Obj = { 'installdir' => $Vsite_NameSpace_Delayed->{$key}->{'installdir'}, 
                                'sqldb' => $Vsite_NameSpace_Delayed->{$key}->{'sqldb'}, 
                                'group' => $NEW_VSITE_OBJ_DATA->{'name'}, 
                                'docrootsuffix' => '', 
                                'idirectory' => $Vsite_NameSpace_Delayed->{$key}->{'idirectory'}, 
                                'doUpgrade' => '', 
                                'autoinstall' => $Vsite_NameSpace_Delayed->{$key}->{'autoinstall'}, 
                                'autoupdate' => $Vsite_NameSpace_Delayed->{$key}->{'autoupdate'}, 
                                'doBackupDB' => '', 
                                'doBackupFiles' => '', 
                                'doRestoreDB' => time(), 
                                'doRestoreFiles' => '', 
                                'appname' => $Vsite_NameSpace_Delayed->{$key}->{'appname'}, 
                                'revision' => $Vsite_NameSpace_Delayed->{$key}->{'revision'}, 
                                'doNothing' => time()
                            };

                    # Create or update toplevel 'WebApplications' Object for this WebApp:
                    $ret = &CCE_Tramp('WebApplications', '', { 'group' => $NEW_VSITE_OBJ_DATA->{'name'}, 'appname' => $Vsite_NameSpace_Delayed->{$key}->{'appname'} }, $WA_Obj);
                }

                # Set NameSpace:
                ($ok) = $cce->set($VSITE_OIDS[0], $key, $Vsite_NameSpace_Delayed->{$key});
                if (!$ok) {
                    &help("Failed to update settings of Vsite NameSpace '$key' of Vsite $VsiteObject->{'fqdn'}");
                }
                else {
                    print "[INFO] Configured NameSpace '$key' of Vsite $VsiteObject->{'fqdn'}.\n";
                }
            }

            #
            ### Import MYSQL_Vsite Databases (if there were any):
            #

            if ($TARGETSYSTEM{'System'}{'productBuild'} eq '5210R') {
                $db_restore_dir = $target_Vsite_Basedir . '/wwwroot/sql/';
            }
            else {
                $db_restore_dir = $target_Vsite_Basedir . '/sql/';
            }

            $vsite_DB = $Import_VSITE->{'NameSpaces'}->{'MYSQL_Vsite'}->{'DB'};
            $vsite_DB_file = $db_restore_dir . $Import_VSITE->{'NameSpaces'}->{'MYSQL_Vsite'}->{'DB'} . '.sql';

            if (($Import_VSITE->{'NameSpaces'}->{'MYSQL_Vsite'}->{'DB'}) && (-f $vsite_DB_file)) {
                print "[INFO] Importing '$vsite_DB' SQL database of Vsite " .  $Import_VSITE->{'Vsite'}->{'fqdn'} . "\n";
                #print "Running: /usr/bin/mysql -u " . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_root'} . " -p" . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_rootpassword'} . " " . $vsite_DB . " < $vsite_DB_file\n";
                system("/usr/bin/mysql -u " . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_root'} . " -p" . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_rootpassword'} . " " . $vsite_DB . " < $vsite_DB_file >/dev/null 2>&1 || :");
            }

            @DBmulti = $cce->scalar_to_array($Import_VSITE->{'NameSpaces'}->{'MYSQL_Vsite'}->{'DBmulti'});
            if ((scalar(@DBmulti)) > "0") {
                foreach my $vsite_DB (@DBmulti) {
                    $vsite_DB_file = $db_restore_dir . $vsite_DB . '.sql';
                    if (-f $vsite_DB_file) {
                        print "[INFO] Importing '$vsite_DB' SQL database of Vsite " .  $Import_VSITE->{'Vsite'}->{'fqdn'} . "\n";
                        #print "Running: /usr/bin/mysql -u " . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_root'} . " -p" . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_rootpassword'} . " " . $vsite_DB . " < $vsite_DB_file\n";
                        system("/usr/bin/mysql -u " . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_root'} . " -p" . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_rootpassword'} . " " . $vsite_DB . " < $vsite_DB_file >/dev/null 2>&1 || :");
                    }
                }
            }

            #
            ### Import the User Files:
            #

            &UserFileImport($USERS->{$old_site_name}, $old_site_name);

            #
            ### Handle Subdomains:
            #

            %Vsite_Subdomains = ();
            $Vsite_Subdomains = $Import_VSITE->{'Subdomains'};

            foreach my $key (keys %{ $Vsite_Subdomains }) {
                %Vsite_NameSpace = ();
                $Vsite_NameSpace = $Import_VSITE->{'Subdomains'}->{$key};

                #print "Old Subdomains: " . Dumper(\$Vsite_NameSpace);

                if ($Vsite_NameSpace->{'isUser'} == '1') {
                    print "[INFO] Ignoring User-Subdomain '$Vsite_NameSpace->{'hostname'}' as it will be imported on User import.\n";
                    next;
                }
                else {
                    # Delete baggage:
                    delete $Vsite_NameSpace->{'CLASS'};
                    delete $Vsite_NameSpace->{'CLASSVER'};
                    delete $Vsite_NameSpace->{'OID'};
                    delete $Vsite_NameSpace->{'NAMESPACE'};
                    delete $Vsite_NameSpace->{'group'};
                    delete $Vsite_NameSpace->{'props'};

                    if (($Vsite_NameSpace->{'hostname'} eq '') && ($TARGETSYSTEM{'System'}{'productBuild'} eq '5209R')) {
                        print "[WARN] NOT importing Subdomain '$Vsite_NameSpace->{'domainname'}' as these type of 'Web Server Alias' webspace allocations only work on 5210R.\n";
                        delete $Vsite_NameSpace->{'domainname'};
                        next;
                    }

                    if (($Vsite_NameSpace->{'domainname'}) && ($TARGETSYSTEM{'System'}{'productBuild'} eq '5209R')) {
                        delete $Vsite_NameSpace->{'domainname'};
                    }

                    if ((!$Vsite_NameSpace->{'domainname'}) && ($TARGETSYSTEM{'System'}{'productBuild'} eq '5210R')) {
                        $Vsite_NameSpace->{'domainname'} = $VsiteObject->{'domain'};
                    }

                    # Fix group:
                    $Vsite_NameSpace->{'group'} = $NEW_VSITE_OBJ_DATA->{'name'};

                    # Fix path:
                    $webpathFinal = (split '/', $Vsite_NameSpace->{'webpath'})[-1];
                    $interMezzo = (split '/', $Vsite_NameSpace->{'webpath'})[-2];

                    if ($TARGETSYSTEM{'System'}{'productBuild'} eq '5210R') {
                        if ($interMezzo eq 'vhosts') {
                            $Vsite_NameSpace->{'webpath'} = $NEW_VSITE_OBJ_DATA->{'basedir'} . '/wwwroot/' . $webpathFinal;
                        }
                        else {
                            $Vsite_NameSpace->{'webpath'} = $NEW_VSITE_OBJ_DATA->{'basedir'} . '/wwwroot/web/' . $webpathFinal;
                        }
                    }
                    else {
                        if ($interMezzo eq 'vhosts') {
                            $Vsite_NameSpace->{'webpath'} = $NEW_VSITE_OBJ_DATA->{'basedir'} . '/vhosts/' . $webpathFinal;
                        }
                        else {
                            $Vsite_NameSpace->{'webpath'} = $NEW_VSITE_OBJ_DATA->{'basedir'} . '/web/' . $webpathFinal;
                        }
                    }
                }
                if ($TARGETSYSTEM{'System'}{'productBuild'} eq '5210R') {
                    print "[INFO] Importing Subdomain '$Vsite_NameSpace->{'hostname'}.$Vsite_NameSpace->{'domainname'}'\n";
                    $ret = &CCE_Tramp('Subdomains', '', { 'hostname' => $Vsite_NameSpace->{'hostname'}, 'domainname' => $Vsite_NameSpace->{'domainname'}, 'group' => $Vsite_NameSpace->{'group'} }, $Vsite_NameSpace);
                }
                else {
                    print "[INFO] Importing Subdomain '$Vsite_NameSpace->{'hostname'}'.\n";
                    $ret = &CCE_Tramp('Subdomains', '', { 'hostname' => $Vsite_NameSpace->{'hostname'}, 'group' => $Vsite_NameSpace->{'group'} }, $Vsite_NameSpace);
                }
            }

        } # End of individual Vsite import

        #
        ### Import DNS (if any)
        #

        if ($option_noDNS == '0') {

            # DNS Include file:
            if ($SYSTEM->{'DNS'}->{'DnsIncludeFile'}->{$NEW_VSITE_OBJ_DATA->{'domain'}}) {
                my $include_File = '/var/named/chroot/var/named/db.' . $NEW_VSITE_OBJ_DATA->{'domain'} . '.include';
                print "[INFO] Importing DNS include file for Vsite $NEW_VSITE_OBJ_DATA->{'fqdn'}: $include_File\n";
                system("/usr/bin/rsync -q --progress -ar '-e ssh -p $port' $source:$include_File $include_File --delete-after --info=progress2 --chown=named:named");
            }

            #
            ### Migrate the DNS CODB Objects:
            #

            # Import DnsSOA:
            if ($SYSTEM->{'DNS'}->{'DnsSOA'}->{$NEW_VSITE_OBJ_DATA->{'domain'}}) {
                print "[INFO] Importing DNS records of type 'DnsSOA' for Vsite $NEW_VSITE_OBJ_DATA->{'fqdn'}.\n";
                foreach my $dns_old_oid (keys %{ $SYSTEM->{'DNS'}->{'DnsSOA'}->{$NEW_VSITE_OBJ_DATA->{'domain'}} }) {
                    $record = $SYSTEM->{'DNS'}->{'DnsSOA'}->{$NEW_VSITE_OBJ_DATA->{'domain'}}->{$dns_old_oid};
                    if ($record->{'CLASS'} eq 'DnsSOA') {
                        $ret = &CCE_Tramp($record->{'CLASS'}, '', { 'domainname' => $record->{'domainname'} }, $record);
                        if ($ret == '-1') {
                            print "[WARN] The DNS SOA could not be set properly. Please check the original and imported DNS SOA and adjust it accordingly.\n";
                        }
                    }
                }
            }

            # Import DnsRecord:
            if ($SYSTEM->{'DNS'}->{'DnsRecord'}->{$NEW_VSITE_OBJ_DATA->{'domain'}}) {
                print "[INFO] Importing DNS records of type 'DnsRecord' for Vsite $NEW_VSITE_OBJ_DATA->{'fqdn'}.\n";
                foreach my $dns_old_oid (keys %{ $SYSTEM->{'DNS'}->{'DnsRecord'}->{$NEW_VSITE_OBJ_DATA->{'domain'}} }) {
                    $record = $SYSTEM->{'DNS'}->{'DnsRecord'}->{$NEW_VSITE_OBJ_DATA->{'domain'}}->{$dns_old_oid};
                    if ($record->{'CLASS'} eq 'DnsRecord') {
                        my %record_cleaned = ();
                        my %record_new = ();
                        $record_cleaned = &CCE_Stripper($record);

                        # Note to self: Shallow Hash copy is sufficient here:
                        $record_new = { %$record_cleaned };

                        # Conditionally change IPs in records to new IP:
                        if (($record_new->{'type'} ne 'AAAA') && ($record_new->{'ipaddr'} eq $old_site_ipv4)) {
                            $record_new->{'ipaddr'} = $VsiteObject->{'ipaddr'};
                            $record_new->{'network'} = $VsiteObject->{'ipaddr'} . '/32';
                        }
                        if (($record_new->{'type'} eq 'AAAA') && ($record_new->{'ipaddr'} eq $old_site_ipv6)) {
                            $record_new->{'ipaddr'} = $VsiteObject->{'import_ipv6'};
                        }
                        $ret = &CCE_Tramp('DnsRecord', '', { 'type' => $record->{'type'}, 'hostname' => $record->{'hostname'}, 'domainname' => $record->{'domainname'} }, $record_new);
                        if ($ret == '-1') {
                            print "[WARN] Not all DNS records imported fine. Please check the original and imported DNS records and adjust them accordingly.\n";
                        }
                    }
                }
            }

            # Import DnsSlaveZone:
            if ($SYSTEM->{'DNS'}->{'DnsSlaveZone'}->{$NEW_VSITE_OBJ_DATA->{'domain'}}) {
                print "[INFO] Importing DNS records of type 'DnsSlaveZone' for Vsite $NEW_VSITE_OBJ_DATA->{'fqdn'}.\n";
                foreach my $dns_old_oid (keys %{ $SYSTEM->{'DNS'}->{'DnsSlaveZone'}->{$NEW_VSITE_OBJ_DATA->{'domain'}} }) {
                    $record = $SYSTEM->{'DNS'}->{'DnsSlaveZone'}->{$NEW_VSITE_OBJ_DATA->{'domain'}}->{$dns_old_oid};
                    if ($record->{'CLASS'} eq 'DnsSlaveZone') {
                        $ret = &CCE_Tramp($record->{'CLASS'}, '', { 'domain' => $record->{'domainname'} }, $record);
                        if ($ret == '-1') {
                            print "[WARN] DNS Slave Zone could not be configured. Most likely you do have primary records for this DNS Zone in place.\n";
                        }
                    }
                }
            }
        } # End of $option_noDNS condition

    } # End Vsite import

    print "\n";
}

sub MigrateMySQL {

    $source = shift;
    $port = shift;
    $host = $source . ':' . $port;

    print "[MYSQL_MIGRATION_START] Starting Migration of MySQL users and databases.\n";

    # Directory setup:
    system("mkdir -p /home/.easy-migrate/");
    system("chmod 0700 /home/.easy-migrate/");

    # Trigger MySQL-Dump on Source:
    $ssh->system("/usr/sausalito/bin/em-helper.pl --mysql >/dev/null 2>&1 || :") or die "remote command failed: " . $ssh->error;
    $ssh->system("mkdir -p /home/.easy-migrate/") or die "remote command failed: " . $ssh->error;
    $ssh->system("chmod 0700 /home/.easy-migrate/") or die "remote command failed: " . $ssh->error;

    $sourceDir = '/home/.easy-migrate/';

    # Fetch remote SQL databases:
    system("/usr/bin/rsync -q --progress -ar '-e ssh -p $port' $source:$sourceDir $sourceDir --delete-after --info=progress2 --chown=root:root >/dev/null 2>&1 || :");

    if (-f '/home/.easy-migrate/mysql-all-dbs.sql') {
        system("/usr/bin/mysql -u " . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_root'} . " -p" . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_rootpassword'} . " < /home/.easy-migrate/mysql-all-dbs.sql >/dev/null 2>&1 || :");
    }

    if (-f '/home/.easy-migrate/mysql-all-dbs.sql') {
        #print "/usr/bin/mysql -u " . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_root'} . " -p" . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_rootpassword'} . " < /home/.easy-migrate/mysql-all-dbs.sql \n";
        system("/usr/bin/mysql -u " . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_root'} . " -p" . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_rootpassword'} . " < /home/.easy-migrate/mysql-all-dbs.sql >/dev/null 2>&1 || :");
    }
    if (-f '/home/.easy-migrate/mysqlusergrants.sql') {
        #print "/usr/bin/mysql -u " . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_root'} . " -p" . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_rootpassword'} . " < /home/.easy-migrate/mysqlusergrants.sql \n";
        system("/usr/bin/mysql -u " . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_root'} . " -p" . $TARGETSYSTEM{'System'}{'MySQL'}{'sql_rootpassword'} . " < /home/.easy-migrate/mysqlusergrants.sql >/dev/null 2>&1 || :");
    }

    print "\n[MYSQL_MIGRATION_END] Migration of MySQL users and databases finished.\n";

    if ($cceinit == '1') {
        $cce->bye("SUCCESS");
    }
    #exit(0);
}

sub MigrateDNS {

    print "[DNS_MIGRATION_START] Starting Migration of all DNS records.\n";

    # Get System Object from Source:
    my $CMD = $ssh->capture("/usr/sausalito/bin/em-helper.pl --system");
    @ret = parser('\[SYSTEM\]', $CMD, 1);
    if ($ret[0] eq "1") {
        $SYSTEM = $ret[1];
    }
    else {
        &help("$ret[1]\n");
    }

    foreach my $CURRENT_DOMAIN (keys %{ $SYSTEM->{'DNS'}->{'DnsSOA'} }) {
        if ($CURRENT_DOMAIN eq "0") {
            print "[INFO] Importing DNS records of type 'DnsSOA' for a network.\n";
        }
        else {
            print "[INFO] Importing DNS records of type 'DnsSOA' for Zone $CURRENT_DOMAIN.\n";
        }
        foreach my $dns_old_oid (keys %{ $SYSTEM->{'DNS'}->{'DnsSOA'}->{$CURRENT_DOMAIN} }) {
            $record = $SYSTEM->{'DNS'}->{'DnsSOA'}->{$CURRENT_DOMAIN}->{$dns_old_oid};
            if ($record->{'CLASS'} eq 'DnsSOA') {
                $ret = &CCE_Tramp($record->{'CLASS'}, '', { 'domainname' => $record->{'domainname'} }, $record);
                if ($ret == '-1') {
                    print "[WARN] The DNS SOA could not be set properly. Please check the original and imported DNS SOA and adjust it accordingly.\n";
                }
            }
        }
    }

    # Import DnsRecord:
    foreach my $CURRENT_DOMAIN (keys %{ $SYSTEM->{'DNS'}->{'DnsRecord'} }) {
        if ($SYSTEM->{'DNS'}->{'DnsRecord'}->{$CURRENT_DOMAIN}) {
            print "[INFO] Importing DNS records of type 'DnsRecord' for Zone $CURRENT_DOMAIN.\n";
            foreach my $dns_old_oid (keys %{ $SYSTEM->{'DNS'}->{'DnsRecord'}->{$CURRENT_DOMAIN} }) {
                $record = $SYSTEM->{'DNS'}->{'DnsRecord'}->{$CURRENT_DOMAIN}->{$dns_old_oid};
                if ($record->{'CLASS'} eq 'DnsRecord') {
                    my %record_cleaned = ();
                    my %record_new = ();
                    $record_cleaned = &CCE_Stripper($record);

                    # Note to self: Shallow Hash copy is sufficient here:
                    $record_new = { %$record_cleaned };

                    # Conditionally change IPs in records to new IP:
                    if (($import_ip ne "") && ($import_ip ne '127.0.0.1')) {
                        if ($record_new->{'type'} ne 'AAAA') {
                            $record_new->{'ipaddr'} = $import_ip;
                            $record_new->{'network'} = $import_ip . '/32';
                        }
                    }
                    if ($import_ipv6 ne "") {
                        if ($record_new->{'type'} eq 'AAAA') {
                            $record_new->{'ipaddr'} = $import_ipv6;
                        }
                    }

                    # DNS Include file:
                    if ($SYSTEM->{'DNS'}->{'DnsIncludeFile'}->{$CURRENT_DOMAIN}) {
                        my $include_File = '/var/named/chroot/var/named/db.' . $CURRENT_DOMAIN . '.include';
                        print "[INFO] Importing DNS include file for domain $CURRENT_DOMAIN: $include_File\n";
                        system("/usr/bin/rsync -q --progress -ar '-e ssh -p $port' $source:$include_File $include_File --delete-after --info=progress2 --chown=named:named");
                    }

                    $ret = &CCE_Tramp('DnsRecord', '', { 'type' => $record->{'type'}, 'hostname' => $record->{'hostname'}, 'domainname' => $record->{'domainname'} }, $record_new);
                    if ($ret == '-1') {
                        print "[WARN] Not all DNS records imported fine. Please check the original and imported DNS records and adjust them accordingly.\n";
                    }
                }
            }
        }
    }

    # Import DnsSlaveZone:
    foreach my $CURRENT_DOMAIN (keys %{ $SYSTEM->{'DNS'}->{'DnsSlaveZone'} }) {
        print "[INFO] Importing DNS records of type 'DnsSlaveZone' for Zone $CURRENT_DOMAIN.\n";
        foreach my $dns_old_oid (keys %{ $SYSTEM->{'DNS'}->{'DnsSlaveZone'}->{$CURRENT_DOMAIN} }) {
            $record = $SYSTEM->{'DNS'}->{'DnsSlaveZone'}->{$CURRENT_DOMAIN}->{$dns_old_oid};
            if ($record->{'CLASS'} eq 'DnsSlaveZone') {
                $ret = &CCE_Tramp($record->{'CLASS'}, '', { 'domain' => $record->{'domain'} }, $record);
                if ($ret == '-1') {
                    print "[WARN] DNS Slave Zone could not be configured. Most likely you do have primary records for this DNS Zone in place.\n";
                }
            }
        }
    }
}

sub UserImport {

    my $Import_USERS = shift;
    my $user_group = shift;

    # Get the Baseline first. For that we use the one User that we know will exist already: 'admin'
    $TARGET_USER_BASELINE = $TARGETUSERS->{'users'}->{'admin'}->{'User'};

    # Continue if we DO have Users to import:
    if ((scalar($Import_USERS)) gt '0') {
        if ($user_group ne '') {
            print "[INFO] Importing owner of Vsite $VsiteObject->{'fqdn'}\n";
        }
        else {
            print "[INFO] Importing user of Vsite $VsiteObject->{'fqdn'}\n";
        }

        foreach my $import_username (keys %{ $Import_USERS }) {

            #if ($TARGETUSERS->{'users'}->{$import_username}) {
                #print "From TARGET_USER_BASELINE: " . Dumper($TARGETUSERS);
                #print "From Import_USERS: " . Dumper($Import_USERS->{$import_username});
                #exit;
            #}

            # Get structure of the User's object from the import:
            %UserObject = ();
            $UserObject = $Import_USERS->{$import_username}->{'User'};

            @BaseLineKeys = (keys %$TARGET_USER_BASELINE);
            @UserKeys = (keys %$UserObject);

            # Get the difference between User Baseline and import User keys:
            my @diff = array_diff(@BaseLineKeys, @UserKeys);
            foreach my $x (@diff) {
                print "delete key: $x \n";
                # If the import Vsite has keys that we do NOT have locally, then we delete them here:
                delete $UserObject->{$x};
            }

            #
            # Import User Object Cleanup:
            #
            # Unconditionally delete the following keys:
            #
            # - OID
            # - CLASSVER
            # - CLASS
            # - NAMESPACE
            #

            delete $UserObject->{'OID'};
            delete $UserObject->{'CLASSVER'};
            delete $UserObject->{'CLASS'};
            delete $UserObject->{'NAMESPACE'};

            # Fix which Vsite Group this User now belongs to:
            $UserObject->{'site'} = $user_group;

            #
            ### Actual User creation & population:
            #

            my @USER_OIDS = ();

            (@USER_OIDS) = $cce->find('User', { 'name' => $import_username });
            if ((scalar(@USER_OIDS)) eq '0') {
                # No User with this unique name was found. Good. Then we create one:
                $cce->create("User", $UserObject);
                $USER_OIDS[0] = $cce->oid();

                # Make sure it's there:
                ($ok, $NEW_USER_OBJ_DATA) = $cce->get($USER_OIDS[0]);
                if (!$ok) {
                    &help("User '$import_username' of Vsite $VsiteObject->{'fqdn'} did not create properly.");
                }
                else {
                    $line = sprintf "%-55s  ", "\n[INFO] User '$import_username' created. ";
                    print $line;
                }
            }
            else {
                # User is already present. Which means we need to check if this is a conflict or an update:
                ($ok, $OLD_USER_OBJ_DATA) = $cce->get($USER_OIDS[0]);
                if ($OLD_USER_OBJ_DATA->{'site'} eq $user_group) {
                    # It's the same User and same Vsite. Update it with the new settings from the import:
                    ($ok) = $cce->set($USER_OIDS[0], '', $UserObject);
                    if (!$ok) {
                        &help("Failed to update settings of User '$import_username'.");
                    }
                    else {
                        $line = sprintf "%-55s  ", "\n[INFO] User '$import_username' updated. ";
                        print $line;
                    }
                }
                else {
                    (@CONFLICT_OIDS) = $cce->find('Vsite', { 'name' => $OLD_USER_OBJ_DATA->{'site'} });
                    ($ok, $CONFLICT_VSITE_DATA) = $cce->get($CONFLICT_OIDS[0]);
                    &help("User '$import_username' of importing Vsite $VsiteObject->{'fqdn'} conflicts with a User of the same name from Vsite " . $CONFLICT_VSITE_DATA->{'fqdn'});
                }
            }

            # Directly get the new Vsite Object:
            if (!defined($NEW_USER_OBJ_DATA)) {
                ($ok, $NEW_USER_OBJ_DATA) = $cce->get($USER_OIDS[0]);
            }

            #
            ### Handle Passwords by setting the 'md5_password' we scraped out of /etc/passwd on the source box:
            #

            ($ok, my $bad, my @info) = $cce->set($USER_OIDS[0], '', { 'md5_password' => $UserObject->{'md5_password'} });
            if (!$ok) { 
                print "[WARN] Encountered a problem setting the 'md5_password' for User '$import_username'.\n";
            }

            #
            ### Populate User NameSpaces:
            #

            @missing_us_namespaces = ();

            # Get the Baseline first. For that we use the one User that we know will exist already: 'admin'
            $TARGET_USER_BASELINE_NS = $TARGETUSERS->{'users'}->{'admin'}->{'NameSpaces'};

            foreach my $key (keys %{ $Import_USERS->{$import_username}->{'NameSpaces'} }) {
                %User_NameSpace = ();
                $User_NameSpace = $Import_USERS->{$import_username}->{'NameSpaces'};
                if (!$TARGET_USER_BASELINE_NS->{$key}) {
                    #print "[INFO] User NameSpace '$key' doesn't exist. You might not have the PKG for that installed. Ignoring settings on import.\n";
                    push @missing_us_namespaces, $key;
                    delete $Import_USERS->{$import_username}->{'NameSpaces'}->{$key};
                }
            }

            # Get the cleaned results again into a Hash:
            %User_NameSpace = ();
            $User_NameSpace = $Import_USERS->{$import_username}->{'NameSpaces'};
            $userns_infoline = '0';
            foreach my $key (keys %{ $User_NameSpace }) {
                %User_NameSpace = ();
                $User_NameSpace = $Import_USERS->{$import_username}->{'NameSpaces'}->{$key};

                %Baseline_NameSpace = ();
                $Baseline_NameSpace = $TARGETUSERS->{'users'}->{'admin'}->{'NameSpaces'}->{$key};

                # We got valid Vsite data
                my @BaseLineKeys = (keys %$Baseline_NameSpace);
                my @UserKeys = (keys %$User_NameSpace);

                # Get the difference between Vsite Baseline and import Vsite keys:
                my @diff = array_diff(@BaseLineKeys, @UserKeys);
                foreach my $x (@diff) {
                    # If the import Vsite has NameSpace keys that we do NOT have locally, then we delete them here:
                    delete $User_NameSpace->{$x};
                    print "Diff in NameSpace $key: $x \n"
                }

                # Cleanup:
                delete $User_NameSpace->{'NAMESPACE'};
                delete $User_NameSpace->{'CLASSVER'};
                delete $User_NameSpace->{'CLASS'};
                delete $User_NameSpace->{'OID'};

                #
                # Sense Checks: Some NameSpace's don't set with empty values:
                #

                if ($key eq 'Disk') {
                    if ($User_NameSpace->{'used'} eq "") {
                        $User_NameSpace->{'used'} = '0';
                    }
                    if ($User_NameSpace->{'over_quota'} eq "") {
                        $User_NameSpace->{'over_quota'} = '0';
                    }
                }

                if (($key eq 'Sites') && ($User_NameSpace->{'used'} eq "")) {
                    $User_NameSpace->{'used'} = '0';
                }

                if (($key eq 'Email') && (($User_NameSpace->{'vacationMsgStart'} eq "") || ($User_NameSpace->{'vacationMsgStop'} eq ""))) {
                    delete $User_NameSpace->{'vacationMsgStart'};
                    delete $User_NameSpace->{'vacationMsgStop'};
                }

                if ($key eq 'ImapSync') {
                    if ($User_NameSpace->{'logdelete'} eq "") {
                        delete $User_NameSpace->{'logdelete'};
                    }
                    if ($User_NameSpace->{'target'} eq "") {
                        delete $User_NameSpace->{'target'};
                    }
                    if ($User_NameSpace->{'targetUsername'} eq "") {
                        delete $User_NameSpace->{'targetUsername'};
                    }
                    if ($User_NameSpace->{'source'} eq "") {
                        delete $User_NameSpace->{'source'};
                    }
                    if ($User_NameSpace->{'sourceUsername'} eq "") {
                        delete $User_NameSpace->{'sourceUsername'};
                    }
                    if ($User_NameSpace->{'dosync'} ne '0') {
                        $User_NameSpace->{'dosync'} = '0';
                    }
                }

                if (($key eq 'Shell') && ($User_NameSpace->{'enabled'} == '1')) {
                    # Make sure that the Vsite has shell enabled. Yes, that is an issue!
                    if ($Import_VSITE->{'NameSpaces'}->{'Shell'}->{'enabled'} == '0') {
                        $User_NameSpace->{'enabled'} = '0';
                    }
                }

                # Set NameSpace:
                ($ok) = $cce->set($USER_OIDS[0], $key, $User_NameSpace);
                if (!$ok) {
                    &help("\n[ERROR] Failed to update settings of User NameSpace '$key' of User '$import_username'.");
                }
                else {
                    #print "[INFO] Configured NameSpace '$key' of User '$import_username'.\n";
                    if ($userns_infoline eq '0') {
                        print "[NameSpace] '$key' ";
                        $userns_infoline++;
                    }
                    else {
                        print "'$key' ";
                    }
                }
            }
        }

        print "\n\n";
        if (scalar(@missing_us_namespaces) gt "0") {
            $outMsg = join(", ", @missing_us_namespaces);
            print "[WARN] Several NameSpace's of Users could not be set. You might not have the PKGs for these installed: $outMsg\n";
        }
    } # End of User import
}

sub UserFileImport {

    my $Import_USERS = shift;
    my $user_group = shift;

    #
    ### Import the User Files:
    #

    # Continue if we DO have Users to import:
    if ((scalar($Import_USERS)) gt '0') {

        if ($user_group ne 'users') {
            print "\n[INFO] Importing Files of Users of Vsite $VsiteObject->{'fqdn'}\n";
        }
        else {
            print "\n[INFO] Importing files of owner of Vsite $VsiteObject->{'fqdn'}\n";
        }

        foreach my $import_username (keys %{ $Import_USERS }) {

            # Vsite Group this User now belongs to:
            $new_user_group = $user_group;

            #
            ### Actual User File synchronization:
            #

            my @USER_OIDS = ();
            (@USER_OIDS) = $cce->find('User', { 'name' => $import_username });
            if ((scalar(@USER_OIDS)) eq '0') {
                # No User with this unique name was found. Good. Then we create one:
                print "Did not find User $import_username in CODB on target. \n";
            }
            else {
                # User is already present. Update it with the new settings from the import:

                # Parse /etc/passwd to find out where the User has his home directory:
                $pw = new Unix::PasswdFile "/etc/passwd";
                $userDir = $pw->home($import_username) . '/';
                undef $pw;

                $sourceDir = $Import_USERS->{$import_username}->{'HomeDirectory'} . '/';

                # Actual Rsync:
                print "[INFO] Starting RSYNC from $host:$sourceDir to $userDir \n";
                #print "/usr/bin/rsync -q --dry-run --progress -ar '-e ssh -p $port' $source:$sourceDir $userDir --delete-after --info=progress2 --chown=$import_username:$new_user_group\n";
                system("/usr/bin/rsync -q --progress -ar '-e ssh -p $port' $source:$sourceDir $userDir --delete-after --info=progress2 --chown=$import_username:$new_user_group >/dev/null 2>&1 || :");

                #
                ### Conditionally fix directory permissions based on platform:
                #

                if ($user_group ne 'users') {
                    if ($TARGETSYSTEM{'System'}{'productBuild'} eq '5210R') {
                        system("chmod 00700 $userDir");
                    }
                    else {
                        system("chmod 02771 $userDir");
                    }
                    if (-f "$userDir/.ftpaccess") {
                        system("chown root:root $userDir/.ftpaccess");
                    }
                }

                #
                ### Handle eventual User cronjobs:
                #
                
                if ($Import_USERS->{$import_username}->{'UserCron'}) {
                    foreach my $userCron (keys %{ $Import_USERS->{$import_username}->{'UserCron'} }) {
                        print "[INFO] User has own cronjob. Making copy of $host:$userCron to $userCron \n";
                        #print "/usr/bin/rsync -q --dry-run --progress -ar '-e ssh -p $port' $source:$userCron $userCron --delete-after --info=progress2 --chown=$import_username:$new_user_group\n";
                        system("/usr/bin/rsync -q --progress -ar '-e ssh -p $port' $source:$userCron $userCron --delete-after --info=progress2 --chown=$import_username:$new_user_group >/dev/null 2>&1 || :");
                        system("chmod 0600 $userCron");
                    }
                }
            }
        }
        print "\n";
    } # End of User File import
}

sub CCE_Stripper {

    my $DirtyObject = shift;

    # Clean the pipes before we make steam:
    delete $DirtyObject->{'CLASS'};
    delete $DirtyObject->{'CLASSVER'};
    delete $DirtyObject->{'NAMESPACE'};
    delete $DirtyObject->{'OID'};
    return $DirtyObject;

}

sub CCE_Tramp {

    # A little helper that creates CODB objects or updates existing ones:

    my $CLASS = shift;
    my $NAMESPACE = shift;
    my $QUERY = shift;
    my $PAYLOAD = shift;

    # Clean the pipes before we make steam:
    delete $PAYLOAD->{'CLASS'};
    delete $PAYLOAD->{'CLASSVER'};
    delete $PAYLOAD->{'NAMESPACE'};
    delete $PAYLOAD->{'OID'};

    my @WRK_OIDS = ();

    if ($QUERY eq 'none') {
        (@WRK_OIDS) = $cce->find($CLASS);
    }
    else {
        (@WRK_OIDS) = $cce->find($CLASS, $QUERY);
    }
    if ((scalar(@WRK_OIDS)) eq '0') {
        #print "Tramping a CREATE on " . $CLASS . " - Scalar: " . scalar(@WRK_OIDS) . " Query: " . Dumper(\$Query);
        ($ok) = $cce->create($CLASS, $PAYLOAD);
        $WRK_OIDS[0] = $cce->oid();
        if (!$ok) {
            return '-1';
        }
    }
    else {
        #print "Tramping a SET on OID $WRK_OIDS[0] for " . $CLASS . " - Scalar: " . scalar(@WRK_OIDS) . " Query: " . Dumper(\$Query);
        ($ok) = $cce->set($WRK_OIDS[0], $NAMESPACE, $PAYLOAD);
        if (!$ok) {
            return '-1';
        }
    }
    ($ok, $WRK_DATA) = $cce->get($WRK_OIDS[0]);
    return $WRK_DATA;
}

sub GetObject {
    my $dat_fqdn = shift;
    my $TYPE = shift;
    my $SUBTYPE = shift;

    $item_of_interest = $$TYPE->{$SUBTYPE}{$dat_fqdn};
    if ($item_of_interest =~ /^site(.*)$/ ) {
        (my $num = $item_of_interest) =~ s/site//;
        if ($$TYPE->{$num}->{$item_of_interest}) {
            return $$TYPE->{$num}->{$item_of_interest}
        }
        else {
            return -1;
        }
    }
    return -1;
}

sub GetBaselineObject {
    my $SiteName = shift;
    my $TYPE = shift;
    my $SUBTYPE = shift;

    (my $num = $SiteName) =~ s/site//;
    if ($$TYPE->{$num}->{$SiteName}->{$SUBTYPE}) {
        return $$TYPE->{$num}->{$SiteName}->{$SUBTYPE};
    }
    else {
        return -1;
    }
}

sub GetBaseLine {

    # Init CCE:
    if ($cceinit == '0') {
        &BlueOnyxInit();
    }

    #
    ### Target SYSTEM Baseline:
    #

    # Get System Object:
    ($sysoid) = $cce->find('System');
    ($ok, $System) = $cce->get($sysoid);

    $TARGETSYSTEM = ();

    # Get the most important info from the System Object:
    $TARGETSYSTEM{'System'}{'productBuild'} = $System->{productBuild};
    $TARGETSYSTEM{'System'}{'productName'}  = $System->{productName};
    $TARGETSYSTEM{'System'}{'hostname'}     = $System->{hostname};
    $TARGETSYSTEM{'System'}{'domainname'}   = $System->{domainname};
    $TARGETSYSTEM{'System'}{'FQDN'}         = $System->{hostname} . '.' . $System->{domainname};

    # Get List of all System NameSpaces:
    ($ok, @SystemNameSpaces) = $cce->names($sysoid);

    # Get all System NameSpaces:
    foreach my $NS (@SystemNameSpaces) {
        ($ok, $SNS) = $cce->get($sysoid, $NS);
        $TARGETSYSTEM{'System'}{'NameSpaces'}{$NS}   = $SNS;
    }

    # Get MYSQL Object:
    ($mysqloid) = $cce->find('MySQL');
    ($ok, $MySQL) = $cce->get($mysqloid);
    if ($ok) {
        $TARGETSYSTEM{'System'}{'MySQL'}         = $MySQL;
    }

    # Conditionally get DNS records if DNS is enabled:
    if ($TARGETSYSTEM{'System'}{'NameSpaces'}{'DNS'}{'enabled'} eq '1') {

        # Get DnsSOA:
        (@DnsSOA) = $cce->findx('DnsSOA');
        foreach my $OID (@DnsSOA) {
            ($ok, $dnsObj) = $cce->get($OID);
            $TARGETSYSTEM{'DNS'}{'DnsSOA'}{$dnsObj->{domainname}}   = $dnsObj;
        }

        # Get DnsRecord:
        (@DnsSOA) = $cce->findx('DnsRecord');
        foreach my $OID (@DnsSOA) {
            ($ok, $dnsObj) = $cce->get($OID);
            $TARGETSYSTEM{'DNS'}{'DnsRecord'}{$dnsObj->{domainname}}   = $dnsObj;
        }

        # Get DnsSlaveZone:
        (@DnsSOA) = $cce->findx('DnsSlaveZone');
        foreach my $OID (@DnsSOA) {
            ($ok, $dnsObj) = $cce->get($OID);
            $TARGETSYSTEM{'DNS'}{'DnsSlaveZone'}{$dnsObj->{domain}}   = $dnsObj;
        }
    }

    # Normalize output:
    $TARGETSYSTEM = decode_json(encode_json(\%TARGETSYSTEM));

    # Deal with NetworkPooling:
    ($ok, $Network) = $cce->get($System->{'OID'}, 'Network');
    if ($Network->{'pooling'} eq '1') {
        (@POOL_OIDS) = $cce->find('IPPoolingRange', { 'min' => '127.0.0.1', 'max' => '127.0.0.1' });
        if (!scalar(@POOL_OIDS)) {
            ($ok) = $cce->create('IPPoolingRange', { 'min' => '127.0.0.1', 'max' => '127.0.0.1', 'creation_time' => time(), 'admin' => '&admin&' });
        }
        if ($import_ip ne '') {
            (@POOL_OIDS) = $cce->find('IPPoolingRange', { 'min' => $import_ip, 'max' => $import_ip });
            if (!scalar(@POOL_OIDS)) {
                ($ok) = $cce->create('IPPoolingRange', { 'min' => $import_ip, 'max' => $import_ip, 'creation_time' => time(), 'admin' => '&admin&' });
            }
        }
        if ($import_ipv6 ne '') {
            (@POOL_OIDS) = $cce->find('IPPoolingRange', { 'min' => $import_ipv6, 'max' => $import_ipv6 });
            if (!scalar(@POOL_OIDS)) {
                ($ok) = $cce->create('IPPoolingRange', { 'min' => $import_ipv6, 'max' => $import_ipv6, 'creation_time' => time(), 'admin' => '&admin&' });
            }
        }
    }

    #
    ### Target VSITES Baseline:
    #

    # Get Vsite data:
    %TARGETVSITES = ();
    (@VsiteOIDs) = $cce->findx('Vsite');

    my $created_dummy_vsite = 0;
    if (scalar(@VsiteOIDs) eq "0") {
        # We do not have *any* Vsites yet! So we create one and delete it in order to get a baseline.
        $cce->create("Vsite", {
            'hostname' => 'dummy',
            'domain' => 'blueonyx.it',
            'fqdn' => 'dummy.blueonyx.it',
            'ipaddr'=> '127.0.0.1',
            'createdUser' => 'admin',
            'volume' => '/home'
        });
        # Store OID of created Vsite:
        $created_dummy_vsite = $cce->oid();
    }

    (@VsiteOIDs) = $cce->findx('Vsite');
    foreach my $OID (@VsiteOIDs) {
        ($ok, $vsiteObj) = $cce->get($OID);
        if ($ok) {
            # Makes sorting easier:
            (my $num = $vsiteObj->{name}) =~ s/site//;
            $TARGETVSITES{$num}{$vsiteObj->{name}}{'Vsite'} = $vsiteObj;

            # Create an Index that associates $vsiteObj->{fqdn} with $vsiteObj->{name}:
            $TARGETVSITES{'INDEX'}{$vsiteObj->{fqdn}} = $vsiteObj->{name};

            # Get List of all Vsite NameSpaces:
            ($ok, @VsiteNameSpaces) = $cce->names($OID);

            # Get all Vsite NameSpaces:
            foreach my $NS (@VsiteNameSpaces) {
                ($ok, $SNS) = $cce->get($OID, $NS);
                $TARGETVSITES{$num}{$vsiteObj->{name}}{'NameSpaces'}{$NS}   = $SNS;
            }
        }
    }

    # Normalize output:
    $TARGETVSITES = decode_json(encode_json(\%TARGETVSITES));

    # If we created a Dummy Vsite, then we delete it again now:
    if (($created_dummy_vsite ne 0) && ($created_dummy_vsite ne '-1')) {
        $cce->destroy($created_dummy_vsite);
    }

    #
    ### Target USERS Baseline:
    #

    # Note to self: We have at least 'admin', so we don't need to create/delete a dummy user on a naked target.

    # Get User data:
    %TARGETUSERS = ();
    (@VsiteOIDs) = $cce->findx('User');
    foreach my $OID (@VsiteOIDs) {
        ($ok, $vsiteObj) = $cce->get($OID);
        if ($ok) {
            if ($vsiteObj->{'site'} eq '') {
                $vsiteObj->{'site'} = 'users';
            }

            $TARGETUSERS{$vsiteObj->{'site'}}{$vsiteObj->{'name'}}{'User'} = $vsiteObj;

            # Get List of all User NameSpaces:
            ($ok, @VsiteNameSpaces) = $cce->names($OID);

            # Get all User NameSpaces:
            foreach my $NS (@VsiteNameSpaces) {
                ($ok, $SNS) = $cce->get($OID, $NS);
                $TARGETUSERS{$vsiteObj->{'site'}}{$vsiteObj->{'name'}}{'NameSpaces'}{$NS}   = $SNS;
            }
        }
    }

    # Normalize output:
    $TARGETUSERS = decode_json(encode_json(\%TARGETUSERS));
}

sub parser {
    $EXPR = shift;
    @ARR =  split("\n", shift);
    $decode = shift || 0;

    # Note to self: We're using List::Util 'first' here for the checks, 
    # as that returns the first match and is pretty fast.

    # Script on source reported an error?
    my $match = first { /^ERROR:(.*)$/ } @ARR;
    if ($match) {
        # It did. Strip out the error and return the bad news:
        $match =~ s/^ERROR: //;
        return ("-1", $match);
    }

    # No error from source script. Check if return data has expected result:
    my $match = first { /^$EXPR(.*)$/ } @ARR;
    if ($match) {
        # It has. 

        if ($decode eq "0") {
            # Return success and the match:
            return ("1", $match);
        }
        else {
            # Strip out the $EXPR and decode the JSON and return success and the hash:
            $match =~ s/$EXPR: //;
            my $result = decode_json($match);
            return ("1", $result);
        }
    }
    else { 
        # It doesn't. Return the bad news and an empty match:
        return ("-1", '');
    }
}

sub header {
    print "┌──────────────────────────────────────────────────────────────┐\n";
    print "│ Easy Migrate Helper: Easily Migrate between BlueOnyx Servers │\n";
    print "└──────────────────────────────────────────────────────────────┘\n\n";
    $header_sent++;
}

sub help {
    $error = shift || "";
    if ($header_sent eq '0') {
        &header;
    }
    if ($error) {
        print "ERROR: $error\n\n";
    }
    print "usage:     easy-migrate.pl [OPTIONS]\n";
    print "\n";
    print "Example: ./easy-migrate.pl --source 10.1.250.1 -p 22 --ip 10.1.252.1 --vsites all\n";
    print "\n";
    print "       --source     Hostname or IP of BlueOnyx to migrate from.\n";
    print "       --port       Port that SSH runs on at the source server.\n";
    print "       --check      Test if source platform is suitable as migration source.\n";
    print "       --list       Human readable list of Vsites present on source server.\n";
    print "       --ip         IPv4 IP Address that newly imported Vsites should get on import.\n";
    print "       --ipv6       IPv6 IP Address that newly imported Vsites should get on import.\n";
    print "       --vsites     To import a single Vsite specify  --vsites www.domain.com\n";
    print "                    To import multiple Vsites specify --vsites www.domain1.com,www.domain2.com\n";
    print "       --config     Configures services and system settings of this server the same\n";
    print "                    way as they are on the source server (Email, DNS, Apache, PHP,\n";
    print "                    Nginx, Vsite- and User-defaults, SNMPd, Sitestats and Support).\n";
    print "                    It will NOT touch the SSH or AdmServ settings for security reasons.\n";
    print "       --nodns      If Vsites are imported, it will not import their DNS records.\n";
    print "       --dnsonly    Only copies all DNS records, but nothing else.\n";
    print "       --mysql      Only copies all MySQL databases and users, but nothing else.\n";
    print "       --version    Show Version of this script.\n";
    print "       -h|--help    This help text\n\n";
    if ($cceinit == '1') {
        $cce->bye("SUCCESS");    
    }
    exit(1);
}


sub find_first_free_group {
    for (my $i = 1; $i <= 2 ** 16; $i++) {
        # use getgrnam to check for available groups, 
        # but this creates a possible
        # race condition.  would CCE take care of the race condition?
        if (not getgrnam("site$i")) {
            # found an availble name
            $name = "site$i";
            return $name;
        }
    }
    return -1;
}

sub uniq (@) {
    my %h;
    map { $h{$_}++ == 0 ? $_ : () } @_;
}

sub in_array {
    my ($arr,$search_for) = @_;
    my %items = map {$_ => 1} @$arr; # create a hash out of the array values
    return (exists($items{$search_for}))?1:0;
}

sub root_check {
    my $id = `id -u`;
    chomp($id);
    if ($id ne "0") {
        &help("$0 must be run by user 'root'!");
    }
}

sub debug_msg {
    if ($DEBUG eq "1") {
        $msg = shift;
        setlogsock('unix');
        openlog($0,'','user');
        syslog('info', "$ARGV[0]: $msg");
        closelog;
    }
    if ($DEBUG eq "2") {
        my $msg = shift;
        print $msg;
    }
}

# 
# Copyright (c) 2008-2020 Michael Stauber, SOLARSPEED.NET
# Copyright (c) 2008-2020 Team BlueOnyx, BLUEONYX.IT
# All Rights Reserved.
# 
# 1. Redistributions of source code must retain the above copyright 
#     notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright 
#     notice, this list of conditions and the following disclaimer in 
#     the documentation and/or other materials provided with the 
#     distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its 
#     contributors may be used to endorse or promote products derived 
#     from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
# 
# You acknowledge that this software is not designed or intended for 
# use in the design, construction, operation or maintenance of any 
# nuclear facility.
# 
